from django.core.management.base import BaseCommand
from django.contrib.auth.models import User
from ...bundles import Bundle


class Command(BaseCommand):
    help = "Render metadata.xml"

    def add_arguments(self, parser):
        parser.add_argument('--site-url')
        parser.add_argument('--bundles-root')
        parser.add_argument('--bundle-name')

    def handle(self, *args, **options):
        b = Bundle(
            site_url=options.get('site_url'),
            bundles_root=options.get('bundles_root'),
            name=options.get('bundle_name'),
        )
        self.stdout.write(b.render_metadata())




from django.contrib.auth.models import User
from django.conf import settings


class SamlBackend(object):

    def authenticate(self, request, saml2_auth=None):
        """
        Get or create user from a OneLogin_Saml2_Auth object.
        """
        username = saml2_auth.get_nameid()
        user, _ = User.objects.get_or_create(username=username)
        attrs = saml2_auth.get_attributes()
        if attrs:
            request.session['logon_attributes_token'] = attrs.get('logon_attributes_token', [''])[0]
        return user

    def get_user(self, user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None



import uuid
import xmlsec
import requests
from lxml import etree
from path import Path
from enum import Enum
from datetime import datetime, timedelta

from django.conf import settings
from django.core.urlresolvers import reverse
from django.template.loader import render_to_string

from onelogin.saml2.constants import OneLogin_Saml2_Constants as constants
from onelogin.saml2.utils import OneLogin_Saml2_Utils

from . import app_settings

import logging
log = logging.getLogger(__name__)

URL_TOKEN_ISSUE = 'https://ws.ite.realme.govt.nz/iCMS/Issue_v1_1'

NAMESPACES = {
    'ds': 'http://www.w3.org/2000/09/xmldsig#',
    'ec': "http://www.w3.org/2001/10/xml-exc-c14n#",
    'env': 'http://www.w3.org/2003/05/soap-envelope',
    'soap': 'http://www.w3.org/2003/05/soap-envelope',
    'wsa': "http://www.w3.org/2005/08/addressing",
    'wsp': 'http://schemas.xmlsoap.org/ws/2004/09/policy',
    'wsse': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd',
    'wsse11': 'http://docs.oasis-open.org/wss/oasis-wss-saml-token-profile-1.1#SAMLV2.0',
    'wst': "http://docs.oasis-open.org/ws-sx/ws-trust/200512",
    'wst14': "http://docs.oasis-open.org/ws-sx/ws-trust/200802",
    'wsu': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd',
    'saml2': 'urn:oasis:names:tc:SAML:2.0:assertion',
    'iCMS': "urn:nzl:govt:ict:stds:authn:deployment:igovt:gls:iCMS:1_0",
}

# override the TODOs with `BUNDLES` in your settings.py
BUNDLES_DEFAULT = {
    'MTS': {
        'idp_entity_id': 'https://mts.realme.govt.nz/saml2',
        'saml_idp_cer': 'mts_login_saml_idp.cer',
        'mutual_ssl_idp_cer': 'mts_mutual_ssl_idp.cer',
        'single_sign_on_service': 'https://mts.realme.govt.nz/logon-mts/mtsEntryPoint',
        'seamless_logon_service': 'NA',
        'saml_sp_cer': 'mts_saml_sp.cer',
        'saml_sp_key': 'mts_saml_sp.key',
        'mutual_ssl_sp_cer': 'mts_mutual_ssl_sp.cer',
        'mutual_ssl_sp_key': 'mts_mutual_ssl_sp.key',
        'sp_entity_id': '',  # TODO
        'site_url': '',  # TODO
    },
    'ITE': {
        'idp_entity_id': 'https://www.ite.logon.realme.govt.nz/saml2',
        'saml_idp_cer': 'ite.signing.logon.realme.govt.nz.cer',
        'mutual_ssl_idp_cer': 'ws.ite.realme.govt.nz.cer',
        'single_sign_on_service': 'https://www.ite.logon.realme.govt.nz/sso/logon/metaAlias/logon/logonidp',
        'seamless_logon_service': 'https://www.ite.logon.realme.govt.nz/cls/seamlessEndpoint',
        'saml_sp_cer': '',  # TODO
        'saml_sp_key': '',  # TODO
        'mutual_ssl_sp_cer': '',  # TODO
        'mutual_ssl_sp_key': '',  # TODO
        'sp_entity_id': '',  # TODO
        'site_url': '',  # TODO
    },
    'PRD': {
        'idp_entity_id': 'https://www.logon.realme.govt.nz/saml2',
        'saml_idp_cer': 'signing.logon.realme.govt.nz.cer',
        'mutual_ssl_idp_cer': 'ws.realme.govt.nz.cer',
        'single_sign_on_service': 'https://www.logon.realme.govt.nz/sso/logon/metaAlias/logon/logonidp',
        'seamless_logon_service': 'https://www.logon.realme.govt.nz/cls/seamlessEndpoint',
        'saml_sp_cer': '',  # TODO
        'saml_sp_key': '',  # TODO
        'mutual_ssl_sp_cer': '',  # TODO
        'mutual_ssl_sp_key': '',  # TODO
        'sp_entity_id': '',  # TODO
        'site_url': '',  # TODO
    }
}


def dt_fmt(dt):
    return dt.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'


def pretty_xml(xml):
    element = etree.fromstring(xml)
    return etree.tostring(element, pretty_print=True).decode('utf-8')


def get_file_body(text):
    lines = text.strip().splitlines()
    start = 1 if lines[0].startswith('-----BEGIN ') else 0
    end = -1 if lines[-1].startswith('-----END ') else None
    return ''.join(lines[start:end])


class AuthnContextClassRef(object):
    LowStrength = 'urn:nzl:govt:ict:stds:authn:deployment:GLS:SAML:2.0:ac:classes:LowStrength'
    ModStrength = 'urn:nzl:govt:ict:stds:authn:deployment:GLS:SAML:2.0:ac:classes:ModStrength'
    # SHOULD NOT be used by a integrating Client SP without first obtaining approval from the RealMe Logon Service.
    ModStrength__OTP_Token_SID = 'urn:nzl:govt:ict:stds:authn:deployment:GLS:SAML:2.0:ac:classes:ModStrength::OTP:Token:SID'
    ModStrength__OTP_Mobile_SMS = 'urn:nzl:govt:ict:stds:authn:deployment:GLS:SAML:2.0:ac:classes:ModStrength::OTP:Mobile:SMS'


class AuthStrength(Enum):
    """
    Enumeration of supported RealMe authentication strengths
    """

    low = (10, AuthnContextClassRef.LowStrength)
    moderate = (20, AuthnContextClassRef.ModStrength)

    def __init__(self, rating, authn_context):
        """
        :param rating: The 'Responder Deemed Authentication Strength', given in RealMe specs
        :param authn_context: the default AuthnContextClassRef to use in RealMe authentication
        """
        self.rating = rating
        self.authn_context = authn_context

    @classmethod
    def from_authn_context(cls, authn_context):
        for auth_strength in cls:
            if authn_context == auth_strength.authn_context:
                return auth_strength
        raise ValueError("Unrecognised AuthnContextClassRef '{}'".format(authn_context))


class Bundle(object):

    def __init__(self, site_url=None, bundles_root=None, name=None, healthcheck=True):
        self.name = name or app_settings.BUNDLE_NAME
        self.bundles_final_config = self.get_bundles_final_config()
        self.config = self.bundles_final_config[self.name]

        self.bundles_root = Path(bundles_root or app_settings.BUNDLES_ROOT)
        self.path = self.bundles_root / self.name

        self.site_url = site_url or self.config.get('site_url') or settings.SITE_URL
        if healthcheck:
            self.healthcheck()

    def healthcheck(self):
        assert self.name in self.bundles_final_config, 'invalid bundle name: {}'.format(self.name)
        assert self.bundles_root.isdir(), self.bundles_root
        assert self.path.isdir(), self.path

        fields = (
            'saml_idp_cer',
            'mutual_ssl_idp_cer',
            'saml_sp_cer',
            'saml_sp_key',
            'mutual_ssl_sp_cer',
            'mutual_ssl_sp_key',
        )

        for field in fields:
            if self.config.get(field):
                assert self.file_path(field).isfile()

        for prefix in ('saml_sp', 'mutual_ssl_sp'):
            self.check_cer_and_key(prefix)


    def __str__(self):
        return self.name

    def get_bundles_final_config(self):
        """
        Load `BUNDLES` from settings.py, and merge with `BUNDLES_DEFAULT`.

        The key in `BUNDLES` can have sub name.
        E.g.: You can define 2 ITE environments, ITE-uat and ITE-testing
        """
        final_config = {}
        for key, conf in app_settings.BUNDLES.items():
            env = key.split('-')[0]  # ITE-uat --> ITE
            default_conf = BUNDLES_DEFAULT[env].copy()
            default_conf.update(conf)  # update default with conf
            final_config[key] = default_conf
        return final_config

    def get_target_sp_entity_id(self, target_sp):
        return self.config.get('target_sps', {}).get(target_sp, {}).get('entity_id', '')

    def check_cer_and_key(self, prefix):
        from subprocess import check_output
        cer_field = '{}_cer'.format(prefix)
        key_field = '{}_key'.format(prefix)
        if self.config.get(cer_field) and self.config.get(key_field):
            cer_path = self.file_path(cer_field)
            key_path = self.file_path(key_field)
            cmd = 'openssl x509 -noout -modulus -in {}'.format(cer_path)
            s1 = check_output(cmd.split())
            cmd = 'openssl rsa -noout -modulus -in {}'.format(key_path)
            s2 = check_output(cmd.split())
            assert s1 == s2

    def file_path(self, field):
        return self.path / self.config[field]

    def file_text(self, field):
        return self.file_path(field).text().strip()

    def file_body(self, field):
        return get_file_body(self.file_text(field))

    @property
    def idp_cer_body(self):
        return self.file_body('saml_idp_cer')

    @property
    def sp_cer_body(self):
        return self.file_body('saml_sp_cer')

    @property
    def sp_key_body(self):
        return self.file_body('saml_sp_key')

    def full_url(self, url):
        return self.site_url.strip('/') + url

    @property
    def idp_entity_id(self):
        return self.config['idp_entity_id']

    @property
    def sp_entity_id(self):
        return self.config['sp_entity_id'].strip()

    @property
    def sp_acs_url(self):
        return self.full_url(reverse('realme:acs'))

    def get_settings(self, authn_context=AuthnContextClassRef.LowStrength):
        return {
            "strict": True,
            "debug": settings.DEBUG,
            "security": {
                "nameIdEncrypted": False,
                "authnRequestsSigned": True,
                "requestedAuthnContext": [authn_context],
                "logoutRequestSigned": False,
                "logoutResponseSigned": False,
                "signMetadata": False,
                "wantMessagesSigned": False,
                "wantAssertionsSigned": True,
                "wantAttributeStatement": False,
                "wantNameId": True,
                "wantNameIdEncrypted": False,
                "wantAssertionsEncrypted": False,
                "signatureAlgorithm": constants.RSA_SHA1,
            },
            "sp": {
                "entityId": self.sp_entity_id,
                "assertionConsumerService": {
                    "url": self.sp_acs_url,
                    "binding": constants.BINDING_HTTP_POST,
                },
                # "singleLogoutService": {
                #     "url": "",
                #     "binding": constants.BINDING_HTTP_REDIRECT,
                # },
                "NameIDFormat": constants.NAMEID_UNSPECIFIED,
                "x509cert": self.sp_cer_body,
                "privateKey": self.sp_key_body,
            },
            "idp": {
                "entityId": self.idp_entity_id,
                "singleSignOnService": {
                    "url": self.config['single_sign_on_service'],
                    "binding": constants.BINDING_HTTP_REDIRECT,
                },
                # "singleLogoutService": {
                #     "url": "",
                #     "binding": constants.BINDING_HTTP_REDIRECT,
                # },
                "x509cert": self.idp_cer_body,
            }
        }

    def render_metadata(self):
        return render_to_string(
            'realme/metadata.xml',
            {
                'conf': self,
                'METADATA_ORG_NAME': app_settings.METADATA_ORG_NAME,
                'METADATA_ORG_DISPLAY_NAME': app_settings.METADATA_ORG_DISPLAY_NAME,
                'METADATA_ORG_URL': app_settings.METADATA_ORG_URL,
            }
        )

    @property
    def key_identifier(self):
        fp_ascii = OneLogin_Saml2_Utils.calculate_x509_fingerprint(self.sp_cer_body)
        # --> unicode str in ascii: 'b5a10daa4250aea3b036b4a2a6e66829f852363f'
        from binascii import a2b_hex, b2a_base64
        fp_hex = a2b_hex(fp_ascii)
        # --> unicode str in hex: '\xb5\xa1\r\xaaBP\xae\xa3\xb06\xb4\xa2\xa6\xe6h)\xf8R6?'
        return b2a_base64(fp_hex).strip()  # 'taENqkJQrqOwNrSipuZoKfhSNj8=\n'

    def render_opaque_token_request(self, logon_attributes_token='', target_sp_entity_id=''):
        created = datetime.utcnow()  # must use utc time
        expires = created + timedelta(minutes=5)
        return render_to_string(
            'realme/opaque_token_request_tmpl.xml',
            context={
                'conf': self,
                'created': dt_fmt(created),
                'expires': dt_fmt(expires),
                'message_id': str(uuid.uuid4()),
                'to': URL_TOKEN_ISSUE,
                'REF_IDS': ("Id-Action", "Id-MessageID", "Id-To", "Id-ReplyTo", "Id-Body", "Id-Timestamp"),
                'NAMESPACES': NAMESPACES,
                'key_identifier': self.key_identifier,
                'logon_attributes_token': logon_attributes_token,
                'target_sp_entity_id': target_sp_entity_id,
            }
        )

    def sign_opaque_token_request(self, rendered_xml):
        root_element = etree.fromstring(rendered_xml)
        xmlsec.tree.add_ids(root_element, ["Id"])  # important!
        # refer to xml/opaque_token_request_unsigned.xml for example
        signature_element = root_element.xpath('/soap:Envelope/soap:Header/wsse:Security/ds:Signature', namespaces=NAMESPACES)[0]

        key_path = self.file_path('saml_sp_key')
        assert key_path.isfile
        cer_path = self.file_path('saml_sp_cer')
        assert cer_path.isfile

        # Load private key (assuming that there is no password).
        key = xmlsec.Key.from_file(key_path, xmlsec.KeyFormat.PEM)
        assert key is not None

        # Load the certificate and add it to the key.
        key.load_cert_from_file(cer_path, xmlsec.KeyFormat.PEM)

        # Create a digital signature context (no key manager is needed).
        ctx = xmlsec.SignatureContext()
        ctx.key = key
        # Sign the template.
        ctx.sign(signature_element)
        # return a utf-8 encoded byte str
        # refer to xml/opaque_token_request_signed.xml for example
        return etree.tostring(root_element, pretty_print=True).decode('utf-8')

    def verify_opaque_token_response(self, signed_xml):
        root_element = etree.fromstring(signed_xml)
        xmlsec.tree.add_ids(root_element, ["Id"])  # important!
        signature_element = root_element.xpath('/soap:Envelope/soap:Header/wsse:Security/ds:Signature', namespaces=NAMESPACES)[0]

        cer_path = self.file_path('saml_idp_cer')
        assert cer_path.isfile
        key = xmlsec.Key.from_file(cer_path, xmlsec.KeyFormat.CERT_PEM)
        assert key is not None

        # Create a digital signature context (no key manager is needed).
        ctx = xmlsec.SignatureContext()
        ctx.key = key

        from xmlsec.error import VerificationError
        try:
            # no return value, raise exception if failed
            ctx.verify(signature_element)
        except VerificationError as e:
            log.error('verify_opaque_token_response failed: {}'.format(e))
            return False
        return True

    def send_opaque_token_request(self, user, target_sp):
        logon_attributes_token = user.profile.logon_attributes_token
        target_sp_entity_id = self.get_target_sp_entity_id(target_sp)
        rendered_xml = self.render_opaque_token_request(
            logon_attributes_token=logon_attributes_token,
            target_sp_entity_id=target_sp_entity_id,
        )
        log.debug('opaque token request unsigned: %s', rendered_xml)

        signed_xml = self.sign_opaque_token_request(rendered_xml)
        log.debug('opaque token request: %s', signed_xml)

        headers = {'content-type': 'text/xml'}
        cert = (
            self.file_path('mutual_ssl_sp_cer'),
            self.file_path('mutual_ssl_sp_key'),
        )
        return requests.post(URL_TOKEN_ISSUE, data=signed_xml, headers=headers, cert=cert)




import logging
import re

from django import http
from django.conf import settings
from django.shortcuts import redirect, render
from django.core.exceptions import ImproperlyConfigured
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_GET
from django.contrib.auth import authenticate, REDIRECT_FIELD_NAME, login as auth_login
from django.contrib.auth import views as auth_views
from django.contrib.auth.decorators import login_required, user_passes_test

from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.response import OneLogin_Saml2_Response
from onelogin.saml2.settings import OneLogin_Saml2_Settings
from onelogin.saml2.constants import OneLogin_Saml2_Constants
from onelogin.saml2.xml_utils import OneLogin_Saml2_XML

from .bundles import Bundle, AuthStrength
from . import app_settings

log = logging.getLogger(__name__)


def get_saml2_settings(auth_strength=AuthStrength.low):
    bundle_settings = Bundle().get_settings(authn_context=auth_strength.authn_context)
    return OneLogin_Saml2_Settings(settings=bundle_settings, sp_validation_only=True)


def prepare_django_request(request):
    # If server is behind proxys or balancers use the HTTP_X_FORWARDED fields
    return {
        'https': 'on' if request.is_secure() else 'off',
        'http_host': request.META['HTTP_HOST'],
        'script_name': request.META['PATH_INFO'],
        'server_port': request.META.get('HTTP_X_FORWARDED_PORT') or request.META.get('SERVER_PORT'),
        'get_data': request.GET.copy(),
        # Uncomment if using ADFS as IdP, https://github.com/onelogin/python-saml/pull/144
        # 'lowercase_urlencoding': True,
        'post_data': request.POST.copy()
    }


def init_saml2_auth(request, strength=AuthStrength.low):
    req = prepare_django_request(request)
    return OneLogin_Saml2_Auth(req, get_saml2_settings(strength))


@require_GET
def login(request):
    """
    Log in using the configured RealMe service.
    Configured with the BUNDLE_NAME setting.
    'FAKE' value uses a realme-themed standard Django login.
    Any other uses the matching bundle defined in the BUNDLES setting
    """
    bundle_name = getattr(settings, 'BUNDLE_NAME', 'FAKE')

    if bundle_name == 'FAKE':
        return fake_realme_login(request)
    elif bundle_name not in settings.BUNDLES:
        raise ImproperlyConfigured(
            "No bundle named '{}' defined in settings.BUNDLES".format(bundle_name)
        )

    # the url to return to after login. Default to root url
    target_url = request.GET.get(REDIRECT_FIELD_NAME, '/')

    if 'strength' in request.GET:
        strength_param = request.GET.get('strength').lower()

        # if a valid strength is requested then use that
        try:
            auth_strength = AuthStrength[strength_param]
        except KeyError:
            log.warn("unrecognised realme auth strength '%s'. Defaulting to 'low'", strength_param)
            auth_strength = AuthStrength.low

    else:
        auth_strength = AuthStrength.low

    auth = init_saml2_auth(request, auth_strength)
    url = auth.login(return_to=target_url)

    log.debug('sp login SAMLRequest = %s', auth.get_last_request_xml())
    log.debug('sp login url: %s', url)

    return redirect(url)


def fake_realme_login(request):
    """
    A standard Django login with a realme-themed template.
    """
    return auth_views.login(request, template_name='realme/fake-login.html')


@user_passes_test(lambda user: user.is_superuser)
def metadata(request):
    return http.HttpResponse(Bundle().render_metadata(), content_type='text/plain')


def get_status(dom):
    """
    Gets Status from a Response, including RealMe inner subcode, if present.

    This provides a replacement for OneLogin_Saml2_Utils.get_status(), which only
    returns a subcode if there is no StatusMessage. RealMe errors return both a
    subcode and a sub-code.

    :param dom: The Response as XML
    :type: Document

    :returns: The Status, an array with code, subcode and a message.
    :rtype: dict
    """
    status = {}

    status_entry = OneLogin_Saml2_XML.query(dom, '/samlp:Response/samlp:Status')
    if len(status_entry) != 1:
        raise Exception('Missing valid Status on response')

    code_entry = OneLogin_Saml2_XML.query(dom, '/samlp:Response/samlp:Status/samlp:StatusCode', status_entry[0])
    if len(code_entry) != 1:
        raise Exception('Missing valid Status Code on response')
    code = code_entry[0].values()[0]
    status['code'] = code

    subcode_entry = OneLogin_Saml2_XML.query(dom, '/samlp:Response/samlp:Status/samlp:StatusCode/samlp:StatusCode', status_entry[0])
    if len(subcode_entry) == 1:
        status['subcode'] = subcode_entry[0].values()[0]

    status['msg'] = ''
    message_entry = OneLogin_Saml2_XML.query(dom, '/samlp:Response/samlp:Status/samlp:StatusMessage', status_entry[0])
    if len(message_entry) == 1:
        status['msg'] = message_entry[0].text

    return status


def get_authentication_strength(dom):
    """
    Gets authentication strength from a Response
    :param dom: The Response as XML, with the Assertion decrypted
    :type: Document
    :returns: the authentication strength
    :rtype: AuthStrength
    """
    authn_context_entry = OneLogin_Saml2_XML.query(
        dom, '/samlp:Response/saml:Assertion/saml:AuthnStatement/saml:AuthnContext/saml:AuthnContextClassRef')

    if len(authn_context_entry) != 1:
        raise Exception('Missing authentication context on response')

    authn_context = authn_context_entry[0].text

    log.debug("found AuthnContextClassRef: %s", authn_context)

    return AuthStrength.from_authn_context(authn_context)



@csrf_exempt
def assertion_consumer_service(request):

    if request.method != 'POST':
        return http.HttpResponseBadRequest()

    target_url = request.POST.get('RelayState', '/')
    response = http.HttpResponseRedirect(target_url)

    auth = init_saml2_auth(request)
    saml2_response = OneLogin_Saml2_Response(
        get_saml2_settings(),
        request.POST['SAMLResponse'],
    )

    if saml2_response.encrypted:
        response_document = saml2_response.decrypted_document
    else:
        response_document = saml2_response.document

    log.debug(
        'decoded and decrypted SAMLResponse = %s',
        OneLogin_Saml2_XML.to_string(response_document).decode('utf-8')
    )

    status = get_status(response_document)
    # status example: {code: FOO, msg: BAR}
    code = status.get('code')
    if code != OneLogin_Saml2_Constants.STATUS_SUCCESS:
        log.error('saml response status: {}'.format(status))
        subcode = status.get('subcode', '')
        realme_inner_code = subcode.split(':')[-1]
        assert realme_inner_code
        response.set_cookie(
            app_settings.EXCHANGE_COOKIE_NAME,
            realme_inner_code,
            secure=settings.SESSION_COOKIE_SECURE
        )
        return response

    auth.process_response()
    if auth.is_authenticated():
        user = authenticate(saml2_auth=auth)
        if user and user.is_active:
            auth_login(request, user)
            auth_strength = get_authentication_strength(response_document)
            request.session['realme_strength'] = auth_strength.name
            return response

    response.set_cookie(
        app_settings.EXCHANGE_COOKIE_NAME,
        'RealMeError',
        secure=settings.SESSION_COOKIE_SECURE
    )
    return response


def extract_opaque_token(xml):
    """
    Extract raw opaque token from response xml.

    Warning: Must keep the token as it is. Do not use this way:

        from lxml import etree
        elem = etree.fromstring(xml).xpath('//saml2:Assertion', namespaces=NAMESPACES)[0]
        return etree.tostring(elem)

    xpath will add used namespaces to elem, so the token is changed.
    """
    match = re.search(r'<wst:RequestedSecurityToken>(.+?)</wst:RequestedSecurityToken>', xml)
    return match.group(1).strip() if match else ''


def escape_opaque_token(xml):
    """
    Escape raw token xml to put it in form field.

    The RealMe Java backend use this function to escape:

        org.apache.commons.lang.StringEscapeUtils.escapeHtml

    which will escape quote, while most Python libraries don't do this by default.
    So the `quote=True` arg is important.
    """
    from html import escape
    return escape(xml, quote=True)


@login_required
def seamless(request, target_sp=''):
    bundle = Bundle()
    r = bundle.send_opaque_token_request(request.user, target_sp)
    xml = r.content.decode(r.encoding)
    log.debug('opaque token response: %s', xml)
    opaque_token_raw = extract_opaque_token(xml)
    assert opaque_token_raw in xml
    log.debug('opaque token raw: %s', opaque_token_raw)
    opaque_token_escaped = escape_opaque_token(opaque_token_raw)
    log.debug('opaque token escaped: %s', opaque_token_escaped)
    response = render(request, 'realme/seamless.html', context={
        'relay_state': bundle.config.get('target_sps', {}).get(target_sp, {}).get('relay_state', ''),
        'opaque_token': opaque_token_escaped,
        'seamless_logon_service': bundle.config['seamless_logon_service'],
    })
    content = response.content.decode('utf-8')
    assert opaque_token_escaped in content
    log.debug('seamless logon request: %s', content)
    return response



from os.path import abspath, dirname
from urllib.parse import urlsplit

from django.test import TestCase, RequestFactory
from django.test import override_settings
from django.http import QueryDict
from django.conf import settings

from path import Path
from onelogin.saml2.utils import OneLogin_Saml2_Utils

from .views import login
from .bundles import AuthnContextClassRef, Bundle

HERE = Path(dirname(abspath(__file__)))
SAMPLES = HERE/'samples'

class RealMeTestCase(TestCase):

    def test_html_escape(self):
        """
        Compare html escape between Python and Java.
        """
        opaque_token_raw = (SAMPLES/'opaque_token_raw.xml').text()
        from .views import escape_opaque_token
        opaque_token_escaped_python = escape_opaque_token(opaque_token_raw)
        opaque_token_escaped_java = (SAMPLES/'opaque_token_escaped_java.xml').text()
        self.assertEqual(opaque_token_escaped_python, opaque_token_escaped_java)

    def test_bundles_config(self):
        """
        Test customized BUNDLES settings are actually applied.
        """
        for name, conf in settings.BUNDLES.items():
            b = Bundle(name=name, healthcheck=False)  # bundles may not exist
            for key, value in conf.items():
                actual = b.config.get(key)
                self.assertEqual(value, actual, msg='name: {}, key: {}'.format(name, key))


@override_settings(BUNDLE_NAME='MTS')
class LoginViewTests(TestCase):
    """
    Tests for the login view.
    NOTE: These need to be run from inside a project that has the MTS bundle decrypted
    and defined in settings.
    """

    def setUp(self):
        self.factory = RequestFactory(HTTP_HOST='localhost')

    def test_low_strength(self):
        request = self.factory.get('/login/', {'strength': 'low'})
        response = login(request)

        self.assert_redirects_to_mts(response)

        query = self.get_query_dict(response.url)

        self.assertIn('SAMLRequest', query)

        saml_request = self.decode_saml_request(query['SAMLRequest'])

        self.assertIn(
            '<saml:AuthnContextClassRef>{}</saml:AuthnContextClassRef>'.format(AuthnContextClassRef.LowStrength),
            saml_request
        )

    def test_moderate_strength(self):
        request = self.factory.get('/login/', {'strength': 'moderate'})
        response = login(request)

        self.assert_redirects_to_mts(response)

        query = self.get_query_dict(response.url)

        self.assertIn('SAMLRequest', query)

        saml_request = self.decode_saml_request(query['SAMLRequest'])

        self.assertIn(
            '<saml:AuthnContextClassRef>{}</saml:AuthnContextClassRef>'.format(AuthnContextClassRef.ModStrength),
            saml_request
        )

    def test_default_strength(self):
        """
        should default to low strength when the strength parameter isn't provided
        """
        request = self.factory.get('/login/')
        response = login(request)

        self.assert_redirects_to_mts(response)

        query = self.get_query_dict(response.url)

        self.assertIn('SAMLRequest', query)

        saml_request = self.decode_saml_request(query['SAMLRequest'])

        self.assertIn(
            '<saml:AuthnContextClassRef>{}</saml:AuthnContextClassRef>'.format(AuthnContextClassRef.LowStrength),
            saml_request
        )

    def test_invalid_strength(self):
        """
        should default to low strength when the strength parameter isn't a recognized value
        """
        request = self.factory.get('/login/', {'strength': 'high'})
        response = login(request)

        self.assert_redirects_to_mts(response)

        query = self.get_query_dict(response.url)

        self.assertIn('SAMLRequest', query)

        saml_request = self.decode_saml_request(query['SAMLRequest'])

        self.assertIn(
            '<saml:AuthnContextClassRef>{}</saml:AuthnContextClassRef>'.format(AuthnContextClassRef.LowStrength),
            saml_request
        )

    def assert_redirects_to_mts(self, response):
        self.assertEqual(response.status_code, 302)
        url = urlsplit(response.url)
        self.assertEqual(url.netloc, 'mts.realme.govt.nz')
        self.assertEqual(url.path, '/logon-mts/mtsEntryPoint')

    def get_query_dict(self, url):
        query_str = urlsplit(url).query
        return QueryDict(query_str)

    def decode_saml_request(self, encoded_saml_request):
        return str(OneLogin_Saml2_Utils.decode_base64_and_inflate(encoded_saml_request))



import os
from os.path import abspath, dirname, join
from setuptools import find_packages, setup

HERE = abspath(dirname(__file__))
os.chdir(HERE)  # allow setup.py to be run from any path

with open(join(HERE, 'README.md')) as readme:
    README = readme.read()

with open(join(HERE, 'requirements.txt')) as requirements:
    REQUIREMENTS = list(requirements.readlines())

setup(
    name='django-realme',
    version='1.0',
    packages=find_packages(exclude=('samples', 'example')),
    include_package_data=True,
    description='A Django app client for the RealMe service',
    long_description=README,
    url='https://github.com/catalyst/django-realme',
    download_url = 'https://github.com/catalyst/django-realme/archive/1.0.tar.gz',
    author='Catalyst IT',
    author_email='',
    license='GPL',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=REQUIREMENTS,
)



from .bundles import AuthStrength

# this function is not used in this app itself
# however, it was used by downstream code like endoflife.
def realme_auth_strength(request):
    """
    Determine the RealMe authentication strength of this session.
    :param request: a Django request
    :return: the AuthStrength of the RealMe session, or None if not authenticated with RealMe
    """
    try:
        return AuthStrength[request.session.get('realme_strength')]
    except KeyError:
        return None



from django.conf.urls import url
from django.conf import settings
from django.contrib.auth.views import LogoutView

from . import views

urlpatterns = [
    url(r'^metadata/$', views.metadata, name='metadata'),
    url(r'^login/$', views.login, name='login'),
    url(r'^logout/$', LogoutView.as_view(), name='logout'),
    url(r'^acs/$', views.assertion_consumer_service, name='acs'),
    url(r'^seamless/(?P<target_sp>\w+)/$', views.seamless, name='seamless'),
]



from os.path import join
from django.conf import settings

# override `bundels.BUNDLES_DEFAULT` in settings.py
BUNDLES = getattr(settings, 'BUNDLES', {})

# abs path to bundles directory, defaults to `bundles` in repo root
BUNDLES_ROOT = getattr(settings, 'BUNDLES_ROOT', join(settings.BASE_DIR, 'bundles'))

# One of the keys in Bundles, or FAKE
BUNDLE_NAME = getattr(settings, 'BUNDLE_NAME', 'FAKE')

# store realme error in this cookie for frontend to use
EXCHANGE_COOKIE_NAME = getattr(settings, 'EXCHANGE_COOKIE_NAME', 'realme_auth')

# Organization info used to render metadata.xml
METADATA_ORG_NAME = getattr(settings, 'METADATA_ORG_NAME', 'Catalyst')
METADATA_ORG_DISPLAY_NAME = getattr(settings, 'METADATA_ORG_DISPLAY_NAME', 'Catalyst IT')
METADATA_ORG_URL = getattr(settings, 'METADATA_ORG_URL', 'https://www.catalyst.net.nz')
